
<!doctype html>
<html lang="en">
  <head>
    <meta charset="utf-8">
    <meta name="viewport" content="width=device-width, initial-scale=1, shrink-to-fit=no">
    <meta name="author" content="Georgy Meshcheryakov">
    <title>{ModelName}: Report</title>

    <link href="./css/bootstrap.min.css" rel="stylesheet">
    <link href="./css/semopy.css" rel="stylesheet">
    <script src="./js/bootstrap.min.js"></script>
    <style>nav .nav-link {{
    color: black;
}}
nav .nav-link:hover {{
    color: grey;
}}
nav .nav-link.active {{
    color: grey;
}}</style>


  </head>
  <body>
<br><br>
<div class="container-fluid">
	<div class="row">
	<div class="col-sm-8">
	<h2 style="display:inline">Report: {ModelName}</h2>
	</div>
</div>
<div class="container-fluid" >
  <div class="row">
    <div class="col-sm-2">
     <br>
     <div class="sticky-top">
      <nav class="nav flex-column" >
        <a class="p-1 nav-link active" id="model_m" href="#model">Model information</a>
        <a class="p-1 nav-link" id="estimates_m" href="#estimates">Estimates</a>
        <a class="p-1 nav-link" id="indices_m" href="#indices">Fit indices</a>
	<a class="p-1 nav-link" id="matrices_m" href="#matrices">Matrices</a>
      </nav>
     </div>
    </div>
    <div class="col-sm-9" style="border-left: 1px solid">
     <br>
	<a class="anchor" id="model"></a>
	<h3>Model information</h3>
		<div class="row">
		<div class="col-sm-4">Model class:</div><div class="col-sm-2"><span style="background-color: #f6f8fa; font-family: monospace;">{ModelClass}</span></div>
		</div>
		<div class="row">
		<div class="col-sm-4">Number of samples:</div><div class="col-sm-2">{NSamples}</div>
		</div>
		<div class="row">
		<div class="col-sm-4">Objective function:</div><div class="col-sm-2">{Objective}</div>
		</div>
		<div class="row">
		<div class="col-sm-4">Converged:</div><div class="col-sm-2 text-{ConvColor}">{Converged}</div>
		</div>
		<br>
		Description in <b>semopy</b> syntax:
		<div class="col-sm-9" style="background-color: #f6f8fa;max-width: 100%;width: 100%;border: 1px solid">
		<pre style="font-family: monospace">{ModelDescription}</pre>
		</div>
		<br>
		<div class="accordion" id="accordion1">
		<div class="accordion-item">
	    <h2 class="accordion-header" id="headingOne">
	      <button class="accordion-button collapsed" type="button" data-bs-toggle="collapse" data-bs-target="#collapseOne" aria-expanded="false" aria-controls="collapseOne">
		Variables classification
	      </button>
	    </h2>
	    <div id="collapseOne" class="accordion-collapse collapse" aria-labelledby="headingOne" data-bs-parent="#accordion1">
	      <div class="accordion-body">
		{VariableClassification}
	      </div>
	    </div>
	  </div>
	</div>
	<br>
		{Visualization}
		
	<a class="anchor" id="estimates"></a>
	<h3>Estimates</h3>
	<div style="font-family:monospace">
	{Estimates}
	</div>
	<br>
	<div class="accordion" id="accordion3">
	<div class="accordion-item">
	    <h2 class="accordion-header" id="headingThree">
	      <button class="accordion-button collapsed" type="button" data-bs-toggle="collapse" data-bs-target="#collapseThree" aria-expanded="false" aria-controls="collapseThree">
		semopy tabular output
	      </button>
	    </h2>	
	    <div id="collapseThree" class="accordion-collapse collapse" aria-labelledby="headingThree" data-bs-parent="#accordion3">
	      <div class="accordion-body" style="background-color: #f6f8fa;max-width: 100%;width: 100%">
		<pre style="font-family: monospace">{InspectionTable}</pre>
	      </div>
	    </div>
	  </div>
	</div>
	<br>
	<div class="accordion" id="accordion2">
	<div class="accordion-item">
	    <h2 class="accordion-header" id="headingTwo">
	      <button class="accordion-button collapsed" type="button" data-bs-toggle="collapse" data-bs-target="#collapseTwo" aria-expanded="false" aria-controls="collapseTwo">
		Optimizer report
	      </button>
	    </h2>	
	    <div id="collapseTwo" class="accordion-collapse collapse" aria-labelledby="headingTwo" data-bs-parent="#accordion2">
	      <div class="accordion-body" style="background-color: #f6f8fa;max-width: 100%;width: 100%">
		<pre style="font-family: monospace">{OptimizerReport}</pre>
	      </div>
	    </div>
	  </div>
	</div>
	<a class="anchor" id="indices"></a>
	<h3>Fit indices</h3>
	{FitIndices}
	<a class="anchor" id="matrices"></a>
	<h3>Matrices</h3>
 	{Matrices}
    </div>
   </div>
  </div>
</div>
<footer class="footer">
      <div class="container">
	<br>
	<center>
        <span class="text-muted">Generated by <a href="https://semopy.com"><b>semopy</b></a> v. {Version}.</span>
	</center>
      </div>
    </footer>
</body>
<script>
const menus = [document.getElementById("model_m"), document.getElementById("estimates_m"), document.getElementById("indices_m"), document.getElementById("matrices_m")];
const headers = [document.getElementById("model"), document.getElementById("estimates"), document.getElementById("indices"), document.getElementById("matrices")];
var cur = menus[0];
cur.classList.add('active');
window.addEventListener("scroll", function() {{
  headers.forEach(function(item, index, array) {{
    if (window.scrollY >= (item.offsetTop - 1.05 * item.offsetHeight)) {{
        item = menus[index];
        cur.classList.remove("active");
        cur = item;
        cur.classList.add("active");
  }}
}});
}});
</script>

</html>
