#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""Examples for semopy."""
from . import political_democracy
from . import univariate_regression
from . import univariate_regression_many
from . import multivariate_regression
from . import holzinger39
from . import example_rf
from . import example_article